﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {

  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  using Interpreter;

  /// <summary>
  /// If the top item of the BOOLEAN stack is TRUE then this removes the second item on the EXEC stack, leaving the first
  /// item to be executed. If it is false then it removes the first item, leaving the second to be executed. This is similar to
  /// CODE.IF except that it operates on the EXEC stack. This acts as a NOOP unless there are at least two items on the EXEC stack and
  /// one item on the BOOLEAN stack.
  /// </summary>
  [PushExpression(StackType.Exec, "EXEC.IF")]
  public class ExecIfExpression : StatelessExpression {
    public override bool Eval(IPushInterpreter interpreter) {
      // not enough arguments on stack
      if ((interpreter.BooleanStack.Count == 0) || (interpreter.ExecStack.Count < 2)) return false;

      var condition = interpreter.BooleanStack.Pop();

      if (condition) interpreter.ExecStack.RemoveAt(interpreter.ExecStack.Count - 2);
      else interpreter.ExecStack.RemoveTop();

      return true;
    }
  }

  /// <summary>
  ///     Inserts beneath the top item of the EXEC stack a new item of the form
  ///     "( EXEC.Y <TopItem> )".
  /// </summary>
  [PushExpression(StackType.Exec, "EXEC.Y")]
  public class ExecYExpression : StatelessExpression {
    public override bool Eval(IPushInterpreter interpreter) {
      // not enough arguments on stack
      if (interpreter.ExecStack.Count == 0 ||
          interpreter.Configuration.MaxPointsInProgram < 2)
        return false;

      var top = interpreter.ExecStack.Top;
      var execYExpression = ExpressionTable.GetStatelessExpression<ExecYExpression>();
      var result = PushProgram.Create(interpreter.PushProgramPool, top, execYExpression);

      interpreter.ExecStack.SetTop(result);
      interpreter.ExecStack.Add(top);

      return true;
    }
  }

  /// <summary>
  ///     Removes the second item on the EXEC stack.
  /// </summary>
  [PushExpression(StackType.Exec, "EXEC.K")]
  public class ExecKExpression : StatelessExpression {
    public override bool Eval(IPushInterpreter interpreter) {
      if (interpreter.ExecStack.Count < 2) return false;

      var top = interpreter.ExecStack.Pop();
      interpreter.ExecStack.SetTop(top);

      return true;
    }
  }

  /// <summary>
  ///     Pops 3 items from the EXEC stack, which we will call A, B, and C
  ///     (with A being the first one popped). Then pushes a list containing B and C back onto the EXEC stack, followed by
  ///     another instance of C, followed by another instance of A.
  /// </summary>
  [PushExpression(StackType.Exec, "EXEC.S")]
  public class ExecSExpression : StatelessExpression {
    public override bool Eval(IPushInterpreter interpreter) {
      if (interpreter.ExecStack.Count < 3) return false;

      var expression = interpreter.ExecStack.Pop(2);
      var a = expression[1];
      var b = expression[0];
      var c = interpreter.ExecStack.Top;

      var newTop = PushProgram.Create(interpreter.PushProgramPool, c, b);

      interpreter.ExecStack.SetTop(newTop);
      interpreter.ExecStack.Push(c, a);

      return true;
    }
  }
}