﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System;

  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  using Interpreter;

  /// <summary>
  ///     Pushes the sum of the top two items.
  /// </summary>
  [PushExpression(StackType.Integer, "INTEGER.+")]
  public class IntegerAddExpression : PushResultExpression<long> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.IntegerStack, 2, values => values[0] + values[1]);
    }
  }

  /// <summary>
  ///     Pushes the difference of the top two items; that is, the second item minus the top item.
  /// </summary>
  [PushExpression(StackType.Integer, "INTEGER.-")]
  public class IntegerSubtractExpression : PushResultExpression<long> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.IntegerStack, 2, values => values[0] - values[1]);
    }
  }

  /// <summary>
  ///     Pushes the product of the top two items.
  /// </summary>
  [PushExpression(StackType.Integer, "INTEGER.*")]
  public class IntegerMultiplyExpression : PushResultExpression<long> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.IntegerStack, 2, values => values[0] * values[1]);
    }
  }

  /// <summary>
  ///     Pushes the quotient of the top two items; that is, the second item divided by the top item.
  ///     If the top item is zero this acts as a NOOP.
  /// </summary>
  [PushExpression(StackType.Integer, "INTEGER./")]
  public class IntegerDivideExpression : PushResultExpression<long> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.IntegerStack, 2, values => values[0] / values[1], 0);
    }
  }

  /// <summary>
  ///     Pushes the second stack item modulo the top stack item. If the top item is zero this acts as a NOOP. The modulus is
  ///     computed as the
  ///     remainder of the quotient, where the quotient has first been truncated toward negative infinity. (This is taken
  ///     from the definition
  ///     for the generic MOD function in Common Lisp, which is described for example at
  ///     http://www.lispworks.com/reference/HyperSpec/Body/f_mod_r.htm.)
  /// </summary>
  [PushExpression(StackType.Integer, "INTEGER.%")]
  public class IntegerModuloExpression : PushResultExpression<long> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.IntegerStack, 2, values => values[0] % values[1], 0);
    }
  }

  /// <summary>
  ///     Pushes the minimum of the top two items.
  /// </summary>
  [PushExpression(StackType.Integer, "INTEGER.MIN")]
  public class IntegerMinExpression : PushResultExpression<long> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.IntegerStack, 2, values => Math.Min(values[0], values[1]));
    }
  }

  /// <summary>
  ///     Pushes the maximum of the top two items.
  /// </summary>
  [PushExpression(StackType.Integer, "INTEGER.MAX")]
  public class IntegerMaxExpression : PushResultExpression<long> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.IntegerStack, 2, values => Math.Max(values[0], values[1]));
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is less than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackType.Integer, "INTEGER.<")]
  public class IntegerSmallerThanExpression : PushResultExpression<long> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.IntegerStack, interpreter.BooleanStack, 2, values => values[0] < values[1]);
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is greater than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackType.Integer, "INTEGER.>")]
  public class IntegerGreaterThanExpression : PushResultExpression<long> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.IntegerStack, interpreter.BooleanStack, 2, values => values[0] > values[1]);
    }
  }

  /// <summary>
  ///     Pushes 1 if the top BOOLEAN is TRUE, or 0 if the top BOOLEAN is FALSE.
  /// </summary>
  [PushExpression(StackType.Integer, "INTEGER.FROMBOOLEAN")]
  public class IntegerFromBooleanExpression : StatelessExpression {
    public override bool Eval(IPushInterpreter interpreter) {
      if (interpreter.BooleanStack.Count == 0) return false;

      var condition = interpreter.BooleanStack.Pop();
      var value = condition ? 1 : 0;

      interpreter.IntegerStack.Push(value);

      return true;
    }
  }

  /// <summary>
  ///     Pushes the result of truncating the top FLOAT.
  /// </summary>
  [PushExpression(StackType.Integer, "INTEGER.FROMFLOAT")]
  public class IntegerFromFloatExpression : StatelessExpression {
    public override bool Eval(IPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count == 0) return false;

      var value = (int)interpreter.FloatStack.Pop();

      interpreter.IntegerStack.Push(value);

      return true;
    }
  }
}