﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System.Diagnostics;

  using HeuristicLab.Problems.ProgramSynthesis.Push.Data.Pool;

  using Interpreter;

  [Serializable]
  public sealed class PushProgram : Expression {
    private static readonly Expression[] EmptyExpressions = new Expression[0];
    public static readonly PushProgram Empty = new PushProgram();

    private const string Prefix = "(";
    private const string PrefixReduced = "[";
    private const string Postfix = ")";
    private const string PostfixReduced = "]";
    private const string Delimiter = " ";

    private Expression[] expressions;

    public PushProgram() : this(EmptyExpressions) {
    }

    public PushProgram(Expression[] expressions) {
      this.expressions = expressions;
    }

    public bool IsEmpty { get { return Count == 0; } }
    public int Count { get { return expressions.Length; } }

    public IReadOnlyList<Expression> Expressions { get { return this.expressions; } }

    public static PushProgram Create(IManagedPool<PushProgram> pool, params Expression[] expressions) {
      var program = pool.Get();

      program.expressions = expressions;
      program.Reset();

      return program;
    }

    private void Reset() {
      stringRepresentation = null;
      depth = null;
      treeIndex = null;
      hashCode = null;
    }

    public static PushProgram Merge(IManagedPool<PushProgram> pool, PushProgram first, Expression second) {
      var program = pool.Get();

      program.Reset();
      program.expressions = new Expression[first.Expressions.Count + 1];
      first.CopyExpressionsTo(program.expressions);
      program.expressions[0] = first;

      return program;
    }

    public static PushProgram Merge(IManagedPool<PushProgram> pool, Expression first, PushProgram second) {
      var program = pool.Get();

      program.Reset();

      program.expressions = new Expression[second.Expressions.Count + 1];
      program.expressions[0] = first;
      second.CopyExpressionsTo(program.expressions, 0, 1, second.Count);

      return program;
    }

    public static PushProgram Merge(IManagedPool<PushProgram> pool, PushProgram first, PushProgram second) {
      var program = pool.Get();

      program.Reset();
      program.expressions = new Expression[first.Count + second.Count];

      first.CopyExpressionsTo(program.expressions);
      second.CopyExpressionsTo(program.expressions, 0, first.Count, second.Count);

      return program;
    }

    public int IndexOf(Expression expression) {
      return Array.IndexOf(this.expressions, expression);
    }

    private string stringRepresentation;
    public override string StringRepresentation
    {
      get
      {
        if (stringRepresentation == null) stringRepresentation = BuildString();
        return stringRepresentation;
      }
    }


    private int? depth;
    public int Depth
    {
      get
      {
        if (depth == null) depth = CalcDepth();
        return depth.Value;
      }
    }

    private int[] treeIndex;
    private int[] TreeIndex
    {
      get
      {
        if (treeIndex == null) treeIndex = BuildTreeIndex();
        return treeIndex;
      }
    }

    private int? hashCode;
    public override int GetHashCode() {
      if (hashCode == null) hashCode = HashExpressions();
      return hashCode.Value;
    }

    //public PushProgram Copy() {
    //  // as the empty list is a static stateless expression, no copy required
    //  return ReferenceEquals(this, Empty) || this.expressions.IsEmpty
    //    ? Empty
    //    : new PushProgram(this.expressions.Copy());
    //}

    public override bool Eval(IPushInterpreter interpreter) {
      interpreter.ExecStack.Push(this.expressions);
      return true;
    }

    public override string ToString() {
      return StringRepresentation;
    }

    /// <summary>
    /// Returns the amount of elements in the whole tree
    /// </summary>
    /// <returns></returns>
    public int TotalCount
    {
      get
      {
        return this.IsEmpty
            ? 1
            // + 1 because "this" is also counted
            : TreeIndex[0] + 1;
      }
    }

    public IEnumerable<Expression> DepthFirst() {
      foreach (var expr in this.expressions) {
        if (expr.IsProgram) {
          var expandExpr = (PushProgram)expr;

          foreach (var sub in expandExpr.DepthFirst())
            yield return sub;
        } else yield return expr;
      }
    }

    public PushProgram Copy() {
      if (IsEmpty) return this;

      var program = new PushProgram(this.expressions) {
        stringRepresentation = this.stringRepresentation,
        hashCode = this.hashCode,
        depth = this.depth,
        treeIndex = this.treeIndex
      };

      return program;
    }

    public PushProgram Copy(IManagedPool<PushProgram> pool) {
      if (IsEmpty) return this;

      var program = pool.Get();
      program.expressions = this.expressions;
      program.stringRepresentation = stringRepresentation;
      program.hashCode = hashCode;
      program.depth = depth;
      program.treeIndex = treeIndex;

      return program;
    }

    public PushProgram Copy(IManagedPool<PushProgram> pool, int startIndex, int length) {
      if (IsEmpty) return this;

      var program = pool.Get();
      program.expressions = CopyExpressions(startIndex, length);

      program.Reset();

      return program;
    }

    public Expression[] CopyExpressions() {
      return CopyExpressions(Count);
    }

    public Expression[] CopyExpressions(int length) {
      if (IsEmpty) return EmptyExpressions;

      var copy = new Expression[length];
      this.expressions.CopyTo(copy, 0);

      return copy;
    }

    public Expression[] CopyExpressions(int startIndex, int length) {
      if (IsEmpty) return EmptyExpressions;

      var copy = new Expression[length];
      CopyExpressionsTo(copy, startIndex, 0, length);

      return copy;
    }

    public void CopyExpressionsTo(Expression[] destination, int? sourceIndex = null, int? destinationIndex = null, int? length = null) {
      if (IsEmpty) return;

      Array.Copy(this.expressions, sourceIndex ?? 0, destination, destinationIndex ?? 0, length ?? Count);
    }

    private int CalcDepth() {
      var maxDepth = 1;
      for (var i = 0; i < Count; i++) {
        var expression = this.expressions[i];
        if (!expression.IsProgram) continue;
        var expandExpression = (PushProgram)expression;

        if (expandExpression.Depth > maxDepth)
          maxDepth = expandExpression.Depth;
      }

      return maxDepth + 1;
    }

    private string BuildString() {
      // prevent too big string
      return this.TotalCount > 50
        ? PrefixReduced + this.Count + PostfixReduced
        : Prefix + string.Join(Delimiter, this.expressions.Where(e => !string.IsNullOrWhiteSpace(e.StringRepresentation))) + Postfix;
    }

    private int[] BuildTreeIndex() {
      var local_treeIndex = new int[Count];

      var next = 1;

      for (var i = Count - 1; i >= 0; i--) {
        var subExpression = this.expressions[i];

        local_treeIndex[i] = next;

        if (subExpression.IsProgram) {
          next += ((PushProgram)subExpression).TotalCount;
        } else {
          next++;
        }
      }

      return local_treeIndex;
    }

    private int HashExpressions() {
      var hash = 19 * 31 + this.GetType().FullName.GetHashCode();

      for (var i = 0; i < Count; i++) {
        hash = hash * 31 + this.expressions[i].GetHashCode();
      }

      return hash;
    }

    private static void CheckIfNested(int level, PushProgram target, PushProgram current,
        IList<PushProgram> visited) {
      if (visited.Any(x => ReferenceEquals(x, current)))
        Debugger.Break();

      visited.Add(current);
      if (level <= 0) {
        // hirarchy depth > level
        Debugger.Break();
        return;
      }

      for (var i = 0; i < current.Count; i++) {
        if (current.expressions[i].IsProgram)
          continue;

        var subExpression = current.expressions[i] as PushProgram;

        if (ReferenceEquals(target, subExpression)) {
          // Endless recursion dedected
          Debugger.Break();
        }

        CheckIfNested(level - 1, target, subExpression, visited);
      }

      visited.RemoveAt(visited.Count - 1);
    }

    /// <summary>
    ///     Returns the element with the given index whereby the tree is indexed in depth first manner
    /// </summary>
    /// <param name="index">The index of the required element</param>
    /// <returns>The required element</returns>
    public Expression GetFromTree(int index) {
      return GetFromTree(index, (root, parent, child, childIndex, parentIndex) => child);
    }

    /// <returns>The required element</returns>
    public T GetFromTree<T>(int index,
        Func<PushProgram, PushProgram, Expression, int, int, T> resolver) {
      return GetFromTree(index, 0, null, resolver);
    }

    private T GetFromTree<T>(int index, int parentIndex, PushProgram parent,
        Func<PushProgram, PushProgram, Expression, int, int, T> resolver) {
      if (index == 0) return resolver(parent, parent, this, 0, parentIndex);

      var min = Count - 1;
      var max = 0;
      var mid = (min + max) / 2;

      while ((max <= min) && (TreeIndex[mid] != index)) {
        if (TreeIndex[mid] > index) max = mid + 1;
        else min = mid - 1;

        mid = (min + max) / 2;
      }

      return TreeIndex[mid] == index
          ? resolver(parent, this, this.expressions[mid], mid, parentIndex)
          : (this.expressions[mid + 1] as PushProgram).GetFromTree(index - TreeIndex[mid + 1], mid + 1,
              this, resolver);
    }
  }
}
